
#    This file is part of ADLER.
#
#    ADLER is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
# 
# Copyright (C) Maciej Bartkowiak, 2019-2020

from PyQt5.QtGui import QDoubleValidator, QIntValidator
from PyQt5.QtCore import  QObject, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QLabel,  QGroupBox, QGridLayout, QLineEdit, \
                                                QWidget,  QHBoxLayout,  QSizePolicy
import numpy as np
import copy

class MyDoubleValidator(QDoubleValidator):
    def __init__(self, mins, maxes, flen, linedit):
        super().__init__( mins, maxes, flen, linedit)
    def fixup(self, s):
        try:
            val = float(s)
        except:
            return str((self.top() + self.bottom())/2.0)
        else:
            if val > self.top():
                return str(self.top())
            elif val < self.bottom():
                return str(self.bottom())
            else:
                return str((self.top() + self.bottom())/2.0)

class MyIntValidator(QIntValidator):
    def __init__(self, mins, maxes, linedit):
        super().__init__( mins, maxes, linedit)
    def fixup(self, s):
        try:
            val = int(s)
        except:
            return str((self.top() + self.bottom())/2.0)
        else:
            if val > self.top():
                return str(self.top())
            elif val < self.bottom():
                return str(self.bottom())
            else:
                return str((self.top() + self.bottom())/2.0)

class InputVariable(QObject):
    values_changed = pyqtSignal()
    def __init__(self, parent, variable,  grid = None,  prec=9):
        super().__init__(parent)
        self.prec = prec
        self.v_name = variable['Name']
        self.v_unit = variable['Unit']
        self.v_len = variable['Length']
        self.v_value = variable['Value']
        self.v_minval = variable['MinValue']
        self.v_maxval = variable['MaxValue']
        self.v_type = variable['Type']
        self.tooltip=""
        for k in variable.keys():
            self.tooltip += str(k) + ": " + str(variable[k]) + '\n'
        self.maxismatched = False
        self.minismatched = False
        self.isinteger = False
        if 'int' in self.v_type:
            self.isinteger = True
        self.veclen = -1
        self.minlen = -2
        self.maxlen = -3
        self.value = np.zeros(self.v_len)
        self.minval = np.zeros(self.value.shape)
        self.maxval = np.zeros(self.value.shape)
        veclen = self.v_len
        try:
            maxlen = len(self.v_maxval)
        except:
            maxlen = 1
        try:
            minlen = len(self.v_minval)
        except:
            minlen = 1
        self.veclen = veclen
        self.maxlen = maxlen
        self.minlen = minlen
        if maxlen >= veclen:
            self.maxismatched = True
        if minlen >= veclen:
            self.minismatched = True
        if self.v_len > 1:
            for n,  x in enumerate(self.v_value):
                self.value[n] = x
                try:
                    self.maxval[n] = np.array(self.v_maxval)[n]
                except:
                    try:
                        self.maxval[n] = np.array(self.v_maxval)[0]
                    except:
                        self.maxval[n] = 1e12
                try:
                    self.minval[n] = np.array(self.v_minval)[n]
                except:
                    try:
                        self.minval[n] = np.array(self.v_minval)[0]
                    except:
                        self.minval[n] = 1e12
        else:
            self.value[0] = float(self.v_value)
            self.maxval[0] = float(self.v_maxval)
            self.minval[0] = float(self.v_minval)
        if self.isinteger:
            self.value = self.value.astype(np.int)
            self.maxval = self.maxval.astype(np.int)
            self.minval = self.minval.astype(np.int)
        base = QWidget(parent)
        label = QLabel(self.v_name, base)
        label2 = QLabel(self.v_unit, base)
        label.setToolTip(self.tooltip)
        label2.setToolTip(self.tooltip)
        fields = []
        fbase = QWidget(base)
        flay = QHBoxLayout(fbase)
        for ln in range(self.veclen):
            inlin = QLineEdit(str(self.value[ln]), base)
            if self.isinteger:
                dval = MyIntValidator(self.minval[ln], self.maxval[ln], inlin)
            else:
                dval = MyDoubleValidator(self.minval[ln], self.maxval[ln], self.prec, inlin)
            inlin.setValidator(dval)
            inlin.setMinimumHeight(20)
            inlin.setText(str(self.value[ln]))
            inlin.setToolTip(self.tooltip)
            if self.isinteger:
                inlin.returnPressed.connect(self.updateInt)
                inlin.editingFinished.connect(self.updateInt)
            else:
                inlin.returnPressed.connect(self.update)
                inlin.editingFinished.connect(self.update)
            flay.addWidget(inlin)
            fields.append(inlin)
        self.fields = fields
        self.defvalue = copy.deepcopy(self.value)
        if grid is None:
            layout = QHBoxLayout(base)
            # infield.setSizePolicy(QSizePolicy.Preferred,QSizePolicy.Preferred)
            layout.addWidget(label)
            layout.addWidget(fbase)
            layout.addWidget(label2)
        else:
            nGridRows = grid.rowCount()
            grid.addWidget(label,  nGridRows,  0)
            grid.addWidget(fbase,  nGridRows,  1)
            grid.addWidget(label2,  nGridRows,  2)
        if self.isinteger:
            self.updateInt()
        else:
            self.update()
    @pyqtSlot()
    def update(self):
        for n,  tfield in enumerate(self.fields):
            val = tfield.text()
            try:
                val = float(val)
            except:
                val = self.defvalue[n]
                tfield.setText(str(val))
#            else:
#                if val > self.maxval[n]:
#                    val = self.maxval[n]
#                    tfield.setText(str(val))
#                elif val < self.minval[n]:
#                    val = self.minval[n]
#                    tfield.setText(str(val))
            self.value[n] = val
        # print("update was called")
        self.values_changed.emit()
    @pyqtSlot()
    def updateInt(self):
        for n,  tfield in enumerate(self.fields):
            val = tfield.text()
            try:
                val = int(float(val))
            except:
                val = self.defvalue[n]
                tfield.setText(str(val))
#            else:
#                if val > self.maxval[n]:
#                    val = self.maxval[n]
#                    tfield.setText(str(val))
#                elif val < self.minval[n]:
#                    val = self.minval[n]
#                    tfield.setText(str(val))
            self.value[n] = val
        # print("updateInt was called")
        self.values_changed.emit()
    def returnValue(self):
        return self.value
    def takeValue(self,  newvalue):
        for nn in range(len(newvalue)):
            try:
                tval = float(newvalue[nn])
            except:
                continue
            if self.isinteger:
                tval = int(tval)
            if tval > self.maxval[nn]:
                tval = self.maxval[nn]
            if tval < self.minval[nn]:
                tval = self.minval[nn]
            self.value[nn] = tval
            # print(self.v_name, "about to set the ",  nn,  "field to",  tval)
            self.fields[nn].setText(str(round(tval, self.prec)))

class VarBox(QObject):
    values_changed = pyqtSignal()
    def __init__(self, parent, setup_variables,  gname,  prec_override = 9):
        super().__init__(parent)
        self.base = QGroupBox(gname,  parent)
        self.base.setSizePolicy(QSizePolicy.Maximum,QSizePolicy.Maximum)
        self.glay = QGridLayout(self.base)
        self.sample_vars = {}
        self.var_names = []
        self.value_dict = {}
        for v in setup_variables:
            self.sample_vars[v['Key']] = InputVariable(self.base, v,  self.glay,  prec = prec_override)
            self.var_names.append(v['Key'])
            self.value_dict[v['Key']] = self.sample_vars[v['Key']].returnValue()
            self.sample_vars[v['Key']].values_changed.connect(self.updateValues)
    @pyqtSlot()
    def updateValues(self):
        for vn in self.var_names:
            self.value_dict[vn] = self.sample_vars[vn].returnValue()
        # print("SampleBox just did an updateValues")
        self.values_changed.emit()
    @pyqtSlot()
    def updateOutputs(self):
        for vn in self.var_names:
            self.sample_vars[vn].takeValue(self.value_dict[vn])
        # print("SampleBox just did an updateValues")
        self.values_changed.emit()
    def returnValues(self):
        return (self.var_names,  self.value_dict)
    @pyqtSlot(object)
    def receiveValues(self, input):
        value_dict = input
        for vn in value_dict.keys():
            newkey = str(vn).split('-')[-1].split()[0].lower()
            # print("New Key is",  newkey, " old key is",  vn,  " value at the old key",  value_dict[vn])
            for altkey in self.var_names:
                if newkey in altkey:
                    self.sample_vars[altkey].takeValue(value_dict[vn])
        self.updateValues()
        self.updateOutputs()
        self.values_changed.emit()
    def takeValues(self,  var_names, value_dict):
        for vn in var_names:
            if vn in self.var_names:
                self.sample_vars[vn].takeValue(value_dict[vn])
            
